# Web::Application
# Copyright(c) 2002 MoonWolf <moonwolf@moonwolf.com>

module Web
  class Application
    
    class MethodChain
      def initialize
        @chain = []
        @request = nil
        @response = nil
        @session = nil
      end
      attr_accessor :option, :request, :response, :session
      
      def req
        @request
      end
      
      def req=(obj)
        @request = obj
      end

      def rsp
        @response
      end
      
      def rsp=(obj)
        @response = obj
      end
      
      def push(label, m)
        @chain.push [label, m]
      end

      def shift(label=nil)
        @chain.shift
      end

      def unshift(label, m)
        @chain.unshift [label, m]
      end
      
      def index(label)
        catch(:found) {
          @chain.each_with_index {|(l,m),idx|
            if l==label
              throw :found,idx
            end
          }
          nil
        }
      end

      def insert_before(target, label, m)
        if idx=self.index(target)
          @chain[idx,0] = [[label, m]]
        end
      end

      def insert_after(target, label, m)
        if idx=self.index(target)
          @chain[idx+1,0] = [[label, m]]
        end
      end

    end

    def initialize
    end

    def setup(opt={})
      @session_opt = opt['session_opt'] || {}
      @session_key = @session_opt['session_key'] || '_session_id'
      @dbman       = @session_opt['database_manager'] || Web::Persistent::File
      @session_opt['session_key'] = @session_key
      @session_opt['database_manager'] = @dbman
    end

    def teardown
    end

    def start_service(uri, name = self.type.to_s)
      require "drb"
      DRb.start_service()
      ts = DRbObject.new(nil, uri)
      
      loop do
        n, req = ts.in([name, nil])
        break if req == :stop
        rsp = run(req)
        ts.out([req, rsp])
      end
    end
    
    def timeout
      20*60 # sec
    end
    
    def run_mode(req)
      rm = req.get('rm') || 'start'
    end

    def method_chain(request)
      chain = MethodChain.new
      chain.request = request
      chain.response = Web::Response.new
      chain.session = nil

      rm = run_mode(request)
      session = get_session(request)
      chain.session = session
      if session && session['expire']
        chain.push "sequence_incr", :sequence_incr
        
        if expired?(session)
          chain.push "expired",        "expired_" + rm
          chain.push "sessionexpired", :sessionexpired
        else
          chain.push "expire_extend",  :expire_extend
          chain.push "session",        "session_" + rm
        end
      end
      chain.push "do", "do_" + rm
      chain
    end

    def dispatch(chain)
      while m = chain.shift
        label, method = m
        if respond_to?(method)
          begin
            send(method, chain)
          rescue Exception
            error(chain, $!.message + "\n" + $!.backtrace.join("\n"))
          end
        end
      end
      chain.response
    end

    def run(req)
      chain = method_chain(req)
      
      rsp = dispatch(chain)
      
      rsp
    end

    def error(chain, msg)
      rsp = Web::Response.new
      
      msg = h(msg).gsub(/\n/,"<br>")
      rsp.content_type = "text/html"
      rsp.write <<EOS
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN">
<html>
<head>
<title>ERROR #{self.type}</title>
<link rel="INDEX"  href="./">
</head>
<body>
  <h1>ERROR #{self.type}</h1>
  <pre>#{msg}</pre>
</body>
</html>
EOS
      chain.response = rsp
    end

    def get_session(chain)
      session_id = chain.request.cookies[@session_key] || chain.request[@session_key]
      if session_id
        session = Web::Persistent.new(session_id,@session_opt)
        chain.session = session
      else
        nil
      end
    end

    def expired?(session)
      if expire=session['expire']
        if Time.now.to_i >= expire.to_i
          true
        else
          false
        end
      else
        true
      end
    end

    def new_session(chain)
      opt = @session_opt
      session = Web::Persistent.new(nil,opt)
      chain.session = session
      session.transaction do
        session['expire']   = (Time.now.to_i + timeout).to_s
        session['sequence'] = rand(1000000).to_s
        session['state']    = 'start'
      end
      session
    end

    def sequence_incr(session, memo, chain)
      session['sequence'] = (session['sequence'].to_i + 1).to_s
    end

    def expire_extend(session, memo, chain)
      session['expire'] = (Time.now.to_i + timeout).to_s
    end

    # Utility
  
    def h(str)
      Web::escapeHTML(str)
    end
    
    def u(str)
      Web::escape(str.to_s)
    end
    
    
    def qs(hash)
      url = '?'
      url << hash.collect {|key,value|
        u(key.to_s) << '=' << u(value.to_s)
      }.sort.join(';')
      h(url)
    end


    def get_param(chain, name, type, must = true)
      req = chain.request
      raise ArgumentError if must && !req.has_key?(name)
      value ,= req[name]
      return nil if !value and !must
      value = value.read if value.respond_to?(:read)
      case type
      when :text
        value
      when :flag
        case value.trim.downcase
        when /\A(?:on|yes|y|true|1)\z/
          true
        when /\A(?:off|no|n|false|0)\z/
          false
        else
          raise ArgumentError
        end
      when :integer
        Integer(value)
      when Array
        raise ArgumentError unless type.include?(value)
        value
      when Hash
        type[value]
      when Regexp
        raise ArgumentError if value !~ type
      end
    end
  end # Application
end # Web
